#ifndef SHADER_H
#define SHADER_H

#include <map>
#include <vector>

#include <GL/glew.h>

#include "GLTB/refptr.h"

#include "GL3Support/glenums.h"
#include "GL3Support/resource.h"
#include "GL3Support/texture.h"

class TiXmlElement; // forward declaration of TinyXML class to avoid including the header here

namespace GL3Support {
	class Shader : public Resource {
	public:
		enum UniformType {
			Bool,
			Int,
			Float,
			Vec2,
			Vec3,
			Vec4,
			BVec2,
			BVec3,
			BVec4,
			IVec2,
			IVec3,
			IVec4,
			Mat2,
			Mat3,
			Mat4,
			Sampler1D,
			Sampler2D,
			Sampler2DRect,
			Sampler3D,
			SamplerCube,
			Sampler1DShadow,
			Sampler2DShadow,
			Image1D,
			Image2D,
			Image2DRect,
			Image3D,
			ImageCube
		};

		struct Uniform {
			std::string name;
			UniformType type;
			int index;

			int integerValues[6];
			bool booleanValues[4];
			float floatValues[16];
			std::string stringValue;

			bool generatedTexture;
			int textureWidth;
			int textureHeight;
			Texture::Type textureType;
			Texture::Format textureFormat;

			// only for internal use:
			bool _requiresUpdate;
			bool _uniformLocationKnown;
			int _uniformLocation;
			gltb::RefPtr<Texture> _texture;
			bool _textureActivated;
		};

		Shader();
		Shader(std::string filename, std::string shaderName);
		virtual ~Shader();

		void loadDefinition(std::string filename, std::string shaderName);
		std::string getDefinitionFileName() { return definitionFileName; };
		std::string getDefinitionShaderName() { return shaderName; };

		void loadVertexShader(std::string filename);
		void loadFragmentShader(std::string filename);

		void setVertexShader(std::string shader);
		void setFragmentShader(std::string shader);

		void setUniform(std::string uniformName, Uniform uniform);
		Uniform getUniform(std::string uniformName);
		bool hasChangedUniforms() { return uniformsChanged; };

		void setOutputBinding(std::string outputName, int bufferIndex);

		void prepare();
		void activate();
		void deactivate();

		void printMissingUniformsWarings(bool warn) { warnAboutMissingUniforms=warn; };
		void printParameterInfo();

		virtual Type getResourceType() { return Resource::ShaderResource; };
		virtual void reload();

		// used internally only - Mesh requires access to shader attribute list
		int _getProgramID() { return programID; };

	private:
		std::vector<std::string> vertexShaderSources;
		std::vector<std::string> vertexShaderSourceFilenames;
		std::vector<std::string> fragmentShaderSources;
		std::vector<std::string> fragmentShaderSourceFilenames;

		bool shaderFromDefinitionFile;
		std::string definitionFileName;
		std::string shaderName;

		bool vertexShaderFromFile;
		bool fragmentShaderFromFile;

		int vertexShaderID;
		int fragmentShaderID;
		int programID;
		bool uniformsChanged;

		std::vector<Uniform> uniforms;
		std::map<std::string, int> outputBindings;

		bool warnAboutMissingUniforms;

		void printFileInfo(std::string which, std::vector<std::string> &filenames, std::vector<std::string> &shaderSources);

		std::string loadShaderDefines(TiXmlElement *shaderDef);
		std::vector<std::string> loadShaderSourceFilenames(TiXmlElement *shaderDef);
		std::vector<std::string> loadShaderSources(std::string preamble, std::vector<std::string> filenames);

		std::string readShaderFile(const char *filename);
	};
}

#endif
