#include <iostream>

#include "GLTB/exception.h"

#include "GL3Support/glhelper.h"
#include "GL3Support/framebuffer.h"

namespace GL3Support {
	Framebuffer::Framebuffer(GLuint handle, int width, int height, std::string name)
	{
		this->handle=handle;
		this->width=width;
		this->height=height;
		this->name=name;
		isActive=false;
		viewportX1=0;
		viewportY1=0;
		viewportX2=width;
		viewportY2=height;

		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,handle);
		glGenRenderbuffersEXT(1,&depthbufferHandle);
		glBindRenderbufferEXT(GL_RENDERBUFFER_EXT,depthbufferHandle);
		glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT,GL_DEPTH_COMPONENT24,width,height);
		glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,GL_DEPTH_ATTACHMENT_EXT,GL_RENDERBUFFER_EXT,depthbufferHandle);
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);
	}

	Framebuffer::~Framebuffer()
	{
		glDeleteFramebuffersEXT(1,&handle);
	}

	void Framebuffer::addAttachmentToActiveBuffers(GLenum attachment)
	{
		for(unsigned int i=0;i<activatedBuffers.size();i++) {
			if(activatedBuffers[i]==attachment) {
				return;
			}
		}
		activatedBuffers.push_back(attachment);
	}

	void Framebuffer::attachTexture(Attachment attachment, gltb::RefPtr<Texture> texture, Texture::CubeSide cubeSide)
	{
		if(texture->getWidth()!=width || texture->getHeight()!=height) {
			throw gltb::Exception("texture is not attachable because of wrong size","GL3Support::Framebuffer::attachTexture()");
		}

		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,handle);

		switch(attachment) {
		case ColorAttachment:
			switch(texture->getType()) {
			case Texture::TwoD:
				glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT0_EXT,GL_TEXTURE_2D,texture->getTextureID(),0);
				addAttachmentToActiveBuffers(GL_COLOR_ATTACHMENT0_EXT);
				break;
			case Texture::TwoDRect:
				glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT0_EXT,GL_TEXTURE_RECTANGLE_ARB,texture->getTextureID(),0);
				addAttachmentToActiveBuffers(GL_COLOR_ATTACHMENT0_EXT);
				break;
			case Texture::CubeMap:
				glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT0_EXT,cubeSide,texture->getTextureID(),0);
				addAttachmentToActiveBuffers(GL_COLOR_ATTACHMENT0_EXT);
				break;
			}
			break;
		case DepthAttachment:
			switch(texture->getType()) {
			case Texture::TwoD:
				glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_DEPTH_ATTACHMENT_EXT,GL_TEXTURE_2D,texture->getTextureID(),0);
				break;
			case Texture::TwoDRect:
				glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_DEPTH_ATTACHMENT_EXT,GL_TEXTURE_RECTANGLE_ARB,texture->getTextureID(),0);
				break;
			case Texture::CubeMap:
				glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_DEPTH_ATTACHMENT_EXT,cubeSide,texture->getTextureID(),0);
				break;
			}
			break;
		case StencilAttachment:
			glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_STENCIL_ATTACHMENT_EXT,GL_TEXTURE_RECTANGLE_ARB,texture->getTextureID(),0);
			break;
		}

		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);
	}

	void Framebuffer::attachTexture(Attachment attachment, int index, gltb::RefPtr<Texture> texture, Texture::CubeSide cubeSide)
	{
		if(texture->getWidth()!=width || texture->getHeight()!=height) {
			throw gltb::Exception("texture is not attachable because of wrong size","GL3Support::Framebuffer::attachTexture()");
		}

		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,handle);

		switch(attachment) {
		case ColorAttachment:
			switch(texture->getType()) {
			case Texture::TwoD:
				glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT0_EXT+index,GL_TEXTURE_2D,texture->getTextureID(),0);
				addAttachmentToActiveBuffers(GL_COLOR_ATTACHMENT0_EXT+index);
				break;
			case Texture::TwoDRect:
				glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT0_EXT+index,GL_TEXTURE_RECTANGLE_ARB,texture->getTextureID(),0);
				addAttachmentToActiveBuffers(GL_COLOR_ATTACHMENT0_EXT+index);
				break;
			case Texture::CubeMap:
				glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT0_EXT+index,cubeSide,texture->getTextureID(),0);
				addAttachmentToActiveBuffers(GL_COLOR_ATTACHMENT0_EXT+index);
				break;
			}
			break;
		default:
			throw gltb::Exception("textures can only be attached to color buffers by index","GL3Support::Framebuffer::attachTexture()");
		}

		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);
	}

	void Framebuffer::attachTextureLayer(Attachment attachment, gltb::RefPtr<Texture> texture, int layer)
	{
		if(texture->getWidth()!=width || texture->getHeight()!=height) {
			throw gltb::Exception("texture is not attachable because of wrong size","GL3Support::Framebuffer::attachTexture()");
		}

		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,handle);

		switch(attachment) {
		case ColorAttachment:
			switch(texture->getType()) {
			case Texture::ThreeD:
				glFramebufferTexture3DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT0_EXT,GL_TEXTURE_3D,texture->getTextureID(),0,layer);
				addAttachmentToActiveBuffers(GL_COLOR_ATTACHMENT0_EXT);
				break;
			default:
				throw gltb::Exception("can only specify layer when attaching 3D textures","GL3Support::Framebuffer::attachTexture()");
			}
			break;
		case DepthAttachment:
			switch(texture->getType()) {
			case Texture::ThreeD:
				glFramebufferTexture3DEXT(GL_FRAMEBUFFER_EXT,GL_DEPTH_ATTACHMENT_EXT,GL_TEXTURE_3D,texture->getTextureID(),0,layer);
				break;
			default:
				throw gltb::Exception("can only specify layer when attaching 3D textures","GL3Support::Framebuffer::attachTexture()");
			}
			break;
		case StencilAttachment:
			glFramebufferTexture3DEXT(GL_FRAMEBUFFER_EXT,GL_STENCIL_ATTACHMENT_EXT,GL_TEXTURE_3D,texture->getTextureID(),0,layer);
			break;
		}

		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);
	}

	void Framebuffer::attachTextureLayer(Attachment attachment, int index, gltb::RefPtr<Texture> texture, int layer)
	{
		if(texture->getWidth()!=width || texture->getHeight()!=height) {
			throw gltb::Exception("texture is not attachable because of wrong size","GL3Support::Framebuffer::attachTexture()");
		}

		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,handle);

		switch(attachment) {
		case ColorAttachment:
			switch(texture->getType()) {
			case Texture::ThreeD:
				glFramebufferTexture3DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT0_EXT+index,GL_TEXTURE_3D,texture->getTextureID(),0,layer);
				addAttachmentToActiveBuffers(GL_COLOR_ATTACHMENT0_EXT+index);
				break;
			default:
				throw gltb::Exception("can only specify layer when attaching 3D textures","GL3Support::Framebuffer::attachTexture()");
			}
			break;
		default:
			throw gltb::Exception("textures can only be attached to color buffers by index","GL3Support::Framebuffer::attachTexture()");
		}

		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);
	}

	bool Framebuffer::isValid()
	{
		if(!isActive) {
			throw gltb::Exception("cannot check an inactive framebuffer for validity","GL3Support::Framebuffer::isValid()");
		}

		GLenum status=glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);

		switch(status) {
		case GL_FRAMEBUFFER_COMPLETE_EXT:
	//		std::cout << "GL3Support::Framebuffer::isValid(): framebuffer is complete" << std::endl;
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
			std::cout << "GL3Support::Framebuffer::isValid(): framebuffer is incomplete: attachment" << std::endl;
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
			std::cout << "GL3Support::Framebuffer::isValid(): framebuffer is incomplete: missing attachment" << std::endl;
			break;
	//		case GL_FRAMEBUFFER_INCOMPLETE_DUPLICATE_ATTACHMENT_EXT:
	//			std::cout << "Render::Framebuffer::isValid(): framebuffer is incomplete: duplicate attachment" << std::endl;
	//			break;
		case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
			std::cout << "GL3Support::Framebuffer::isValid(): framebuffer is incomplete: dimensions" << std::endl;
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
			std::cout << "GL3Support::Framebuffer::isValid(): framebuffer is incomplete: formats" << std::endl;
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
			std::cout << "GL3Support::Framebuffer::isValid(): framebuffer is incomplete: draw buffer" << std::endl;
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
			std::cout << "GL3Support::Framebuffer::isValid(): framebuffer is incomplete: read buffer" << std::endl;
			break;
		case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
			std::cout << "GL3Support::Framebuffer::isValid(): framebuffer is unsupported" << std::endl;
			break;
	//		case GL_FRAMEBUFFER_STATUS_ERROR_EXT:
	//			std::cout << "Render::Framebuffer::isValid(): status error" << std::endl;
	//			break;
		default:
			std::cout << "GL3Support::Framebuffer::isValid(): unknown status " << status << std::endl;
			break;
		}

		if(status!=GL_FRAMEBUFFER_COMPLETE_EXT) {
			return false;
		}

		return true;
	}

	void Framebuffer::setViewport(int x1, int y1, int x2, int y2)
	{
		viewportX1=x1;
		viewportY1=y1;
		viewportX2=x2;
		viewportY2=y2;
		if(isActive) {
			glViewport(viewportX1, viewportY1, viewportX2, viewportY2);
		}
	}

	void Framebuffer::activate()
	{
		isActive=true;
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,handle);
		if(activatedBuffers.size()>1) {
			GLHelper::checkGLError();
			glDrawBuffers(activatedBuffers.size(),&activatedBuffers[0]);
			GLHelper::checkGLError();
		}
		glViewport(viewportX1, viewportY1, viewportX2, viewportY2);
	}

	void Framebuffer::deactivate()
	{
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);
		isActive=false;
	}
}
