#include <iostream>
#include <fstream>
#include <sstream>
#include <cstring>

#include "tinyxml.h"

#include "GLTB/exception.h"

#include "GL3Support/glhelper.h"
#include "GL3Support/resourcemanager.h"
#include "GL3Support/shader.h"

namespace GL3Support {
	Shader::Shader()
	{
		shaderFromDefinitionFile=false;
		vertexShaderSources.push_back("void main() { gl_Position=ftransform(); }");
		vertexShaderFromFile=false;
		fragmentShaderSources.push_back("void main() { gl_FragColor = vec4(1.0,0.0,1.0,1.0); }");
		fragmentShaderFromFile=false;
		vertexShaderID=-1;
		fragmentShaderID=-1;
		programID=-1;
		uniformsChanged=true;
		warnAboutMissingUniforms=true;
	}

	Shader::Shader(std::string filename, std::string shaderName)
	{
		vertexShaderID=-1;
		fragmentShaderID=-1;
		programID=-1;
		uniformsChanged=true;
		warnAboutMissingUniforms=true;
		try {
			loadDefinition(filename, shaderName);
		} catch(gltb::Error e) {
			throw e;
		} catch(gltb::Exception e) {
			// use absolutely ugly and noticable fallback shaders - violet is a nice color for that
			vertexShaderSources.push_back("void main() { gl_Position=ftransform(); }");
			fragmentShaderSources.push_back("void main() { gl_FragColor = vec4(1.0,0.0,1.0,1.0); }");
			std::cerr << "Exception: " << e.getDescription() << std::endl;
			std::cerr << "In:        " << e.getLocation() << std::endl;
		}
	}

	Shader::~Shader()
	{
		if(programID!=-1) {
			glDeleteShader(fragmentShaderID);
			glDeleteShader(vertexShaderID);
			glDeleteProgram(programID);
		}
	}

	std::string Shader::loadShaderDefines(TiXmlElement *shaderDef)
	{
		std::string defines;

		TiXmlElement *version=shaderDef->FirstChildElement("version");
		if(version!=0) {
			defines+=std::string("#version ") + version->Attribute("value") + "\n";
		}

		TiXmlElement *define=shaderDef->FirstChildElement("define");
		while(define!=0) {
			defines+=std::string("#define ") + define->Attribute("name") + "\t" + define->Attribute("value") + "\n";
			define=define->NextSiblingElement("define");
		}
		return defines;
	}

	std::vector<std::string> Shader::loadShaderSourceFilenames(TiXmlElement *shaderDef)
	{
		TiXmlElement *file=shaderDef->FirstChildElement("file");
		std::vector<std::string> filenames;
		while(file!=0) {
			filenames.push_back(file->Attribute("name"));
			file=file->NextSiblingElement("file");
		}
		return filenames;
	}

	std::vector<std::string> Shader::loadShaderSources(std::string preamble, std::vector<std::string> filenames)
	{
		// load files from filenames and build shader source from them
		std::vector<std::string> sources;

		sources.push_back(preamble);

		for(unsigned int i=0;i<filenames.size();i++) {
			sources.push_back(readShaderFile(filenames[i].c_str()));
		}

		return sources;
	}

	void Shader::loadDefinition(std::string filename, std::string shaderName)
	{
		TiXmlDocument document;

		if(!document.LoadFile(filename.c_str())) {
			throw gltb::Exception("unable to load shader definition file " + filename,"GL3Support::Shader::loadDefinition()");
		}

		TiXmlElement *rootElement=document.RootElement();
		TiXmlElement *shader=rootElement->FirstChildElement("shader");

		bool shaderFound=false;
		do {
			if(shader==0) {
				throw gltb::Exception("shader " + shaderName + " not found in definition file " + filename,"GL3Support::Shader::loadDefinition()");
			}

			if(shaderName==shader->Attribute("name")) {
				// found shader
				shaderFound=true;
				break;
			}

			shader=shader->NextSiblingElement("shader");
		} while(!shaderFound);

		if(!shaderFound) {
			throw gltb::Exception("shader " + shaderName + " not found in definition file " + filename,"GL3Support::Shader::loadDefinition()");
		}

		TiXmlElement *vertexShaderDef=shader->FirstChildElement("vertex");
		std::string vertexShaderDefines=loadShaderDefines(vertexShaderDef);
		vertexShaderSourceFilenames=loadShaderSourceFilenames(vertexShaderDef);
		vertexShaderSources=loadShaderSources(vertexShaderDefines,vertexShaderSourceFilenames);

		TiXmlElement *fragmentShaderDef=shader->FirstChildElement("fragment");
		std::string fragmentShaderDefines=loadShaderDefines(fragmentShaderDef);
		fragmentShaderSourceFilenames=loadShaderSourceFilenames(fragmentShaderDef);
		fragmentShaderSources=loadShaderSources(fragmentShaderDefines,fragmentShaderSourceFilenames);

		TiXmlElement *bindOutput=fragmentShaderDef->FirstChildElement("bind-output");
		while(bindOutput!=0) {
			setOutputBinding(bindOutput->Attribute("variable"),atoi(bindOutput->Attribute("buffer")));
			bindOutput=bindOutput->NextSiblingElement("bind-output");
		}

		shaderFromDefinitionFile=true;
		definitionFileName=filename;
		this->shaderName=shaderName;

		prepare();
	}

	void Shader::loadVertexShader(std::string filename)
	{
		vertexShaderSources.clear();
		vertexShaderSources.push_back(readShaderFile(filename.c_str()));
		vertexShaderFromFile=true;
		vertexShaderSourceFilenames.clear();
		vertexShaderSourceFilenames.push_back(filename);
	}

	void Shader::loadFragmentShader(std::string filename)
	{
		fragmentShaderSources.clear();
		fragmentShaderSources.push_back(readShaderFile(filename.c_str()));
		fragmentShaderFromFile=true;
		fragmentShaderSourceFilenames.clear();
		fragmentShaderSourceFilenames.push_back(filename);
	}

	void Shader::setVertexShader(std::string shader)
	{
		vertexShaderSources.clear();
		vertexShaderSources.push_back(shader);
		vertexShaderFromFile=false;
		vertexShaderSourceFilenames.clear();
	}

	void Shader::setFragmentShader(std::string shader)
	{
		fragmentShaderSources.clear();
		fragmentShaderSources.push_back(shader);
		fragmentShaderFromFile=false;
		fragmentShaderSourceFilenames.clear();
	}

	void Shader::setOutputBinding(std::string outputName, int bufferIndex)
	{
		outputBindings[outputName]=bufferIndex;
	}

	void Shader::printFileInfo(std::string which, std::vector<std::string> &filenames, std::vector<std::string> &shaderSources)
	{
		std::cout << std::endl;
		std::cout << which << " ";
		std::cout << "file names:" << std::endl;
		int startIndex=0;

		for(unsigned int i=startIndex;i<filenames.size();i++) {
			std::cout << i << "\t" << filenames[i] << std::endl;
		}

		if(shaderFromDefinitionFile) {
			std::cout << "shader is " << definitionFileName << ":" << shaderName << std::endl;
			std::cout << "shader preamble: " << std::endl;
			std::cout << shaderSources[0] << std::endl;
		}
		std::cout << std::endl;
	}

	void Shader::prepare()
	{
		// compile and link shaders and check for errors
		char log[8192];
		int result=GL_TRUE;
		const char **sources;

		programID=glCreateProgram();
		vertexShaderID=glCreateShader(GL_VERTEX_SHADER);
		fragmentShaderID=glCreateShader(GL_FRAGMENT_SHADER);

		sources=new const char*[vertexShaderSources.size()];
		for(unsigned int i=0;i<vertexShaderSources.size();i++) {
			sources[i]=vertexShaderSources[i].c_str();
		}

		glShaderSource(vertexShaderID,vertexShaderSources.size(),sources,NULL);
		delete[] sources;
		glCompileShader(vertexShaderID);
		memset(log,0,8192);
		int logLength=-1;
		glGetShaderInfoLog(vertexShaderID,8192,&logLength,log);
		if(logLength>0) {
			std::cout << "vertex shader info log:" << std::endl;
			std::cout << log << std::endl;
			printFileInfo("vertex shader",vertexShaderSourceFilenames, vertexShaderSources);
		}
		glGetShaderiv(vertexShaderID,GL_COMPILE_STATUS,&result);
		if(result!=GL_TRUE) {
			throw gltb::Exception("vertex shader did not compile without errors","GL3Support::Shader::prepare()");
		}

		sources=new const char*[fragmentShaderSources.size()];
		for(unsigned int i=0;i<fragmentShaderSources.size();i++) {
			sources[i]=fragmentShaderSources[i].c_str();
		}

		glShaderSource(fragmentShaderID,fragmentShaderSources.size(),sources,NULL);
		delete[] sources;
		glCompileShader(fragmentShaderID);
		memset(log,0,8192);
		glGetShaderInfoLog(fragmentShaderID,8192,&logLength,log);
		if(logLength>0) {
			std::cout << "fragment shader info log:" << std::endl;
			std::cout << log << std::endl;
			printFileInfo("fragment shader",fragmentShaderSourceFilenames, fragmentShaderSources);
		}
		glGetShaderiv(fragmentShaderID,GL_COMPILE_STATUS,&result);
		if(result!=GL_TRUE) {
			throw gltb::Exception("fragment shader did not compile without errors","GL3Support::Shader::prepare()");
		}

		glAttachShader(programID,vertexShaderID);
		glAttachShader(programID,fragmentShaderID);

		std::map<std::string, int>::iterator bindingIter;
		for(bindingIter=outputBindings.begin();bindingIter!=outputBindings.end();++bindingIter) {
			glBindFragDataLocation(programID,bindingIter->second,bindingIter->first.c_str());
		}

		glLinkProgram(programID);
		memset(log,0,8192);
		glGetProgramInfoLog(programID,8192,&logLength,log);
		if(logLength>0) {
			std::cout << "linker info log:" << std::endl;
			std::cout << log << std::endl;
			printFileInfo("vertex shader",vertexShaderSourceFilenames, vertexShaderSources);
			printFileInfo("fragment shader",fragmentShaderSourceFilenames, fragmentShaderSources);
		}
		glGetProgramiv(programID,GL_LINK_STATUS,&result);
		if(result!=GL_TRUE) {
			throw gltb::Exception("GLSL program failed to link","GL3Support::Shader::prepare()");
		}

		std::vector<Uniform>::iterator iter;
		for(iter=uniforms.begin();iter!=uniforms.end();++iter) {
			switch(iter->type) {
				case Sampler2D:
				case SamplerCube:
					if(iter->_texture==0) {
						if(iter->generatedTexture) {
							iter->_texture=ResourceManager::getInstance()->createTexture(iter->textureType,iter->textureFormat,iter->textureWidth,iter->textureHeight,iter->stringValue);
						} else {
							iter->_texture=ResourceManager::getInstance()->getTexture(iter->stringValue);
						}
					}
					break;
				default: // just to get rid of tons of compiler warnings
					break;
			}
		}

		uniformsChanged=true;
		for(unsigned int i=0;i<uniforms.size();i++) {
			uniforms[i]._requiresUpdate=true;
			uniforms[i]._uniformLocationKnown=false;
		}

		warnAboutMissingUniforms = true;
	}

	void Shader::activate()
	{
		// set active program
		glUseProgram(programID);

		// set shader parameters
		if(uniformsChanged) {
			gltb::RefPtr<Texture> texture;
			int textureUnit=0;
			std::vector<Uniform>::iterator iter;

			for(iter=uniforms.begin();iter!=uniforms.end();++iter) {
				if(!iter->_requiresUpdate) {
					continue;
				}
				iter->_requiresUpdate=false;

				int uniformIndex;
				if(iter->_uniformLocationKnown) {
					uniformIndex=iter->_uniformLocation;
				} else {
					uniformIndex=glGetUniformLocation(programID,iter->name.c_str());
					iter->_uniformLocation=uniformIndex;
					iter->_uniformLocationKnown=true;
				}

				if(uniformIndex==-1 && warnAboutMissingUniforms) {
					std::cout << "WARNING: unable to find uniform " << iter->name << " in shader " << shaderName << std::endl;
					continue;
				}

//				std::cout << "uniform index for " << iter->name << " is " << uniformIndex << std::endl;
				switch(iter->type) {
					case Bool:
						// TODO
						break;
					case Int:
						glUniform1iv(uniformIndex,1,iter->integerValues);
						break;
					case Float:
						glUniform1fv(uniformIndex,1,iter->floatValues);
						break;
					case Vec2:
						glUniform2fv(uniformIndex,1,iter->floatValues);
						break;
					case Vec3:
						glUniform3fv(uniformIndex,1,iter->floatValues);
						break;
					case Vec4:
						glUniform4fv(uniformIndex,1,iter->floatValues);
						break;
					case BVec2:
					case BVec3:
					case BVec4:
						// TODO
						break;
					case IVec2:
						glUniform2iv(uniformIndex,1,iter->integerValues);
						break;
					case IVec3:
						glUniform3iv(uniformIndex,1,iter->integerValues);
						break;
					case IVec4:
						glUniform4iv(uniformIndex,1,iter->integerValues);
						break;
					case Sampler1D:
					case Sampler1DShadow:
					case Sampler2D:
					case Sampler2DRect:
					case Sampler2DShadow:
					case SamplerCube:
					case Image1D:
					case Image2D:
					case Image2DRect:
						if(iter->_texture.getNakedPointer()==0) {
							if(iter->generatedTexture) {
								texture=ResourceManager::getInstance()->createTexture(iter->textureType,iter->textureFormat,iter->textureWidth,iter->textureHeight,iter->stringValue);
							} else {
								texture=ResourceManager::getInstance()->getTexture(iter->stringValue);
							}
							if(texture==0) {
								std::cout << "WARNING: unable to find texture " << iter->stringValue << " which should be bound to shader uniform " << iter->name << std::endl;
							}
							iter->_texture=texture;
						}
						break;
					case Sampler3D:
					case Image3D:
					case ImageCube:
						// TODO
						break;
					case Mat2:
						glUniformMatrix2fv(uniformIndex,1,GL_FALSE,iter->floatValues);
						break;
					case Mat3:
						glUniformMatrix3fv(uniformIndex,1,GL_FALSE,iter->floatValues);
						break;
					case Mat4:
						glUniformMatrix4fv(uniformIndex,1,GL_FALSE,iter->floatValues);
						break;
				}
//				GLHelper::checkGLError();
			}

			uniformsChanged=false;
		}

		// activate textures associated with this shader
		int textureUnit=0;
		int imageUnit=0;
		std::vector<Uniform>::iterator iter;
		for(iter=uniforms.begin();iter!=uniforms.end();++iter) {
			int uniformIndex;
			switch(iter->type) {
				case Sampler1D:
				case Sampler1DShadow:
				case Sampler2D:
				case Sampler2DRect:
				case Sampler2DShadow:
				case SamplerCube:
					if(iter->_texture!=0) {
						if(iter->_uniformLocationKnown) {
							uniformIndex=iter->_uniformLocation;
						} else {
							uniformIndex=glGetUniformLocation(programID,iter->name.c_str());
							iter->_uniformLocation=uniformIndex;
							iter->_uniformLocationKnown=true;
						}
						iter->_texture->bind(textureUnit);
						glUniform1iv(uniformIndex,1,&textureUnit);
						textureUnit++;
						iter->_textureActivated=true;
					}
					break;
				case Sampler3D:
					// TODO
					break;
				case Image1D:
				case Image2D:
				case Image2DRect:
					if(iter->_texture!=0) {
						if(iter->_uniformLocationKnown) {
							uniformIndex=iter->_uniformLocation;
						} else {
							uniformIndex=glGetUniformLocation(programID,iter->name.c_str());
							iter->_uniformLocation=uniformIndex;
							iter->_uniformLocationKnown=true;
						}
						glBindImageTextureEXT(imageUnit,iter->_texture->getTextureID(),0,GL_FALSE,0,GL_READ_WRITE,iter->_texture->getGLFormat());
						glUniform1iv(uniformIndex,1,&imageUnit);
						imageUnit++;
					}
					break;
				case Image3D:
				case ImageCube:
					// TODO
					break;
				default:
					break; // do nothing here - other types should aready have been handled
			}
	//		GLHelper::checkGLError();
		}

		warnAboutMissingUniforms = false;
	}

	void Shader::deactivate()
	{
		glUseProgram(0);

		// disable activated texture units
		std::vector<Uniform>::iterator iter;
		for(iter=uniforms.begin();iter!=uniforms.end();++iter) {
			switch(iter->type) {
			case Sampler1D:
			case Sampler2D:
			case Sampler2DRect:
			case Sampler3D:
			case Sampler1DShadow:
			case Sampler2DShadow:
				if(iter->_textureActivated) {
					iter->_texture->unbind();
					iter->_textureActivated=false;
				}
				break;
			default:
				// don't care
				break;
			}
		}
	}

	void Shader::reload()
	{
		if(programID!=-1) {
			glDeleteShader(fragmentShaderID);
			glDeleteShader(vertexShaderID);
			glDeleteProgram(programID);
		}

		if(shaderFromDefinitionFile) {
			std::cout << "recommpiling shader " << shaderName << " from definition file " << definitionFileName << std::endl;
			loadDefinition(definitionFileName,shaderName);
		} else {
			if(vertexShaderFromFile) {
				loadVertexShader(vertexShaderSourceFilenames[0]);
			}
			if(fragmentShaderFromFile) {
				loadFragmentShader(fragmentShaderSourceFilenames[0]);
			}

			std::cout << "recompiling vertex shader " << vertexShaderSourceFilenames[0] << " and fragment shader " << fragmentShaderSourceFilenames[0] << std::endl;

			prepare();
		}
	}

	void Shader::setUniform(std::string uniformName, Uniform uniform)
	{
		std::vector<Uniform>::iterator iter;

		for(iter=uniforms.begin();iter!=uniforms.end();++iter) {
			if(iter->name==uniformName) {
				if(iter->type!=uniform.type) {
					throw gltb::Exception("type mismatch when trying to set shader parameter " + uniformName,"GL3Support::Shader::setParameter()");
				}
				break;
			}
		}

		if(iter==uniforms.end()) {
	//		std::cout << "GL3Support::Shader::setParameter(): adding new parameter to list: " << parameterName << std::endl;
			uniform.name=uniformName;
			uniforms.push_back(uniform);
			iter=--uniforms.end();
		}

		switch(uniform.type) {
			case Bool:
			case BVec2:
			case BVec3:
			case BVec4:
				for(int i=0;i<4;i++) {
					iter->booleanValues[i]=uniform.booleanValues[i];
				}
				break;
			case Int:
			case IVec2:
			case IVec3:
			case IVec4:
				for(int i=0;i<6;i++) {
					iter->integerValues[i]=uniform.integerValues[i];
				}
				break;
			case Float:
			case Vec2:
			case Vec3:
			case Vec4:
			case Mat2:
			case Mat3:
			case Mat4:
				for(int i=0;i<16;i++) {
					iter->floatValues[i]=uniform.floatValues[i];
				}
				break;
			case Sampler1D:
			case Sampler2D:
			case Sampler2DRect:
			case Sampler3D:
			case SamplerCube:
			case Sampler1DShadow:
			case Sampler2DShadow:
				iter->stringValue=uniform.stringValue;
				iter->_texture=0;
				iter->_textureActivated=false;
				break;
		}

		iter->_requiresUpdate=true;
		iter->_uniformLocationKnown=false;

		uniformsChanged=true;
	}

	Shader::Uniform Shader::getUniform(std::string uniformName)
	{
		std::vector<Uniform>::iterator iter;

		for(iter=uniforms.begin();iter!=uniforms.end();++iter) {
			if(iter->name==uniformName) {
				return *iter;
			}
		}
		throw gltb::Exception("parameter " + uniformName + " not found in shader","GL3Support::Shader::parseShaderDefinition()");
	}

	void Shader::printParameterInfo()
	{
		int numUniforms;

		glGetProgramiv(programID,GL_ACTIVE_UNIFORMS,&numUniforms);

		for(int i=0;i<numUniforms;i++) {
			GLenum type;
			GLint size;
			char variableName[256];
			glGetActiveUniform(programID,i,256,NULL,&size,&type,variableName);

			std::cout << "uniform " << i << " name " << variableName << std::endl;
			// TODO: possible types: FLOAT, FLOAT VEC2, FLOAT VEC3, FLOAT VEC4, INT, INT VEC2, INT VEC3, INT VEC4, BOOL, BOOL VEC2, BOOL VEC3, BOOL VEC4, FLOAT MAT2,	FLOAT MAT3, FLOAT MAT4, FLOAT MAT2x3, FLOAT MAT2x4, FLOAT MAT3x2, FLOAT MAT3x4, FLOAT MAT4x2, FLOAT MAT4x3, SAMPLER 1D, SAMPLER 2D,	SAMPLER 3D, SAMPLER CUBE, SAMPLER 1D SHADOW, SAMPLER 2D SHADOW, SAMPLER 1D ARRAY, SAMPLER 2D ARRAY, SAMPLER 1D ARRAY SHADOW, SAMPLER 2D ARRAY SHADOW, SAMPLER CUBE SHADOW, INT SAMPLER 1D, INT SAMPLER 3D, INT SAMPLER CUBE, INT SAMPLER 2D, INT SAMPLER 1D ARRAY, INT SAMPLER 2D ARRAY, UNSIGNED INT, UNSIGNED INT VEC2, UNSIGNED INT VEC3, UNSIGNED INT VEC4, UNSIGNED INT SAMPLER 1D, UNSIGNED INT SAMPLER 2D, UNSIGNED INT SAMPLER 3D, UNSIGNED INT SAMPLER CUBE, UNSIGNED INT SAMPLER 1D ARRAY, UNSIGNED INT SAMPLER 2D ARRAY
		}

		int numAttributes;

		glGetProgramiv(programID,GL_ACTIVE_ATTRIBUTES,&numAttributes);

		for(int i=0;i<numAttributes;i++) {
			GLenum type;
			GLint size;
			char variableName[256];
			glGetActiveAttrib(programID,i,256,NULL,&size,&type,variableName);
			int attributeIndex=glGetAttribLocation(programID,variableName);

			std::cout << "attribute " << i << " name " << variableName << ", index " << attributeIndex << std::endl;
		}
	}

	std::string Shader::readShaderFile(const char *filename)
	{
		std::string shaderString;

		FILE *file=fopen(filename,"r");
		if(file==0) {
			throw gltb::Exception("unable to open shader source file " + std::string(filename),"GL3Support::Shader::readShaderFile()");
		}

		fseek(file,0,SEEK_END);
		size_t length=ftell(file);
		fseek(file,0,SEEK_SET);

		char *shader=new char[length+1];
		memset(shader,0,length+1);

		fread(shader,sizeof(char),length,file);
		fclose(file);

		shader[length]='\0';
		shaderString=std::string(shader);
		delete[] shader;

		return shaderString;
	}
}
