#include <iostream>
#include <string>
#include <cstring>

#include "GLTB/exception.h"

#include "GL3Support/framebuffer.h"
#include "GL3Support/framebuffermanager.h"
#include "GL3Support/glhelper.h"
#include "GL3Support/texture.h"

namespace GL3Support {
	int Texture::activeTextureUnit=0;

	struct TextureFormatDefinition {
		GLuint glFormat;
		GLuint hostFormat;
		GLuint dataType;
		int bytesPerPixel;
	};

	TextureFormatDefinition textureFormatDefinitions[] = {
			{ GL_R32F,					GL_RED,				GL_FLOAT,			1*sizeof(float) }, // FloatR
			{ GL_RG32F,					GL_RG,				GL_FLOAT,			2*sizeof(float) }, // FloatRG
			{ GL_RGB32F,				GL_RGB,				GL_FLOAT,			3*sizeof(float) }, // FloatRGB
			{ GL_RGBA32F,				GL_RGBA,			GL_FLOAT,			4*sizeof(float) }, // FloatRGBA

			{ GL_R16F,					GL_RED,				GL_FLOAT,			1*sizeof(float) }, // HalfR
			{ GL_R16F,					GL_RG,				GL_FLOAT,			2*sizeof(float) }, // HalfRG
			{ GL_R16F,					GL_RGB,				GL_FLOAT,			3*sizeof(float) }, // HalfRGB
			{ GL_R16F,					GL_RGBA,			GL_FLOAT,			4*sizeof(float) }, // HalfRGBA

			{ GL_R8,					GL_RED,				GL_UNSIGNED_BYTE,	1*sizeof(char)  }, // UCharR
			{ GL_R8I,					GL_RED_INTEGER,		GL_UNSIGNED_BYTE,	1*sizeof(char)  }, // UCharRI
			{ GL_RG8,					GL_RG,				GL_UNSIGNED_BYTE,	2*sizeof(char)  }, // UCharRG
			{ GL_RG8I,					GL_RG_INTEGER,		GL_UNSIGNED_BYTE,	2*sizeof(char)  }, // UCharRGI
			{ GL_RGB8,					GL_RGB,				GL_UNSIGNED_BYTE,	3*sizeof(char)  }, // UCharRGB
			{ GL_RGB8I,					GL_RGB_INTEGER,		GL_UNSIGNED_BYTE,	3*sizeof(char)  }, // UCharRGBI
			{ GL_RGBA8,					GL_RGBA,			GL_UNSIGNED_BYTE,	4*sizeof(char)  }, // UCharRGBA
			{ GL_RGBA8I,				GL_RGBA_INTEGER,	GL_UNSIGNED_BYTE,	4*sizeof(char)  }, // UCharRGBAI

			{ GL_RGB8,					GL_BGR,				GL_UNSIGNED_BYTE,	3*sizeof(char)  }, // UCharBGR
			{ GL_RGB8I,					GL_BGR_INTEGER,		GL_UNSIGNED_BYTE,	3*sizeof(char)  }, // UCharBGRI
			{ GL_RGBA8,					GL_BGRA,			GL_UNSIGNED_BYTE,	4*sizeof(char)  }, // UCharBGRA
			{ GL_RGBA8I,				GL_BGRA_INTEGER,	GL_UNSIGNED_BYTE,	4*sizeof(char)  }, // UCharBGRAI

			{ GL_R8,					GL_RED,				GL_BYTE,			1*sizeof(char)  }, // CharR
			{ GL_R8I,					GL_RED_INTEGER,		GL_BYTE,			1*sizeof(char)  }, // CharRI
			{ GL_RG8,					GL_RG,				GL_BYTE,			2*sizeof(char)  }, // CharRG
			{ GL_RG8I,					GL_RG_INTEGER,		GL_BYTE,			2*sizeof(char)  }, // CharRGI
			{ GL_RGB8,					GL_RGB,				GL_BYTE,			3*sizeof(char)  }, // CharRGB
			{ GL_RGB8I,					GL_RGB_INTEGER,		GL_BYTE,			3*sizeof(char)  }, // CharRGBI
			{ GL_RGBA8,					GL_RGBA,			GL_BYTE,			4*sizeof(char)  }, // CharRGBA
			{ GL_RGBA8I,				GL_RGBA_INTEGER,	GL_BYTE,			4*sizeof(char)  }, // CharRGBAI

			{ GL_R32I,					GL_RED_INTEGER,		GL_INT,				1*sizeof(int)  }, // IntRI
			{ GL_RG32I,					GL_RG_INTEGER,		GL_INT,				2*sizeof(int)  }, // IntRGI
			{ GL_RGB32I,				GL_RGB_INTEGER,		GL_INT,				3*sizeof(int)  }, // IntRGBI
			{ GL_RGBA32I,				GL_RGBA_INTEGER,	GL_INT,				4*sizeof(int)  }, // IntRGBAI

			{ GL_R32UI,					GL_RED_INTEGER,		GL_UNSIGNED_INT,	1*sizeof(int)  }, // UIntRI
			{ GL_RG32UI,				GL_RG_INTEGER,		GL_UNSIGNED_INT,	2*sizeof(int)  }, // UIntRGI
			{ GL_RGB32UI,				GL_RGB_INTEGER,		GL_UNSIGNED_INT,	3*sizeof(int)  }, // UIntRGBI
			{ GL_RGBA32UI,				GL_RGBA_INTEGER,	GL_UNSIGNED_INT,	4*sizeof(int)  }, // UIntRGBAI

			{ GL_DEPTH_COMPONENT, 		GL_DEPTH_COMPONENT, GL_FLOAT, 			1*sizeof(float) }, // Detph
			{ GL_DEPTH_COMPONENT32F,	GL_DEPTH_COMPONENT, GL_FLOAT, 			1*sizeof(float) } // DepthFloat32
	};

	Texture::Texture(Type type, Format format, int width, int height, std::string name, void *data)
	: data(0)
	{
		this->type=type;
		this->format=format;
		this->width=width;
		this->height=height;
		this->name=name;

		glGenTextures(1,&textureID);

		switch(type) {
		case TwoD:
			glType=GL_TEXTURE_2D;
			break;
		case TwoDRect:
			glType=GL_TEXTURE_RECTANGLE_ARB;
			break;
		case ThreeD:
			glType=GL_TEXTURE_3D;
			break;
		case CubeMap:
			glType=GL_TEXTURE_CUBE_MAP;
			break;
		case Buffer:
			glType=GL_TEXTURE_BUFFER;
			break;
		}

		glBindTexture(glType,textureID);

		imageSize=width*height;

		glFormat=textureFormatDefinitions[format].glFormat;
		hostFormat=textureFormatDefinitions[format].hostFormat;
		dataType=textureFormatDefinitions[format].dataType;
		imageSize*=textureFormatDefinitions[format].bytesPerPixel;

		glTexParameteri(glType,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
		glTexParameteri(glType,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
		glTexParameteri(glType,GL_TEXTURE_WRAP_S,GL_CLAMP);
		glTexParameteri(glType,GL_TEXTURE_WRAP_T,GL_CLAMP);

		if(type!=CubeMap && type!=Buffer) {
			GLHelper::checkGLError();
			std::cout << "glTexImage2D()" << std::endl;
    		glTexImage2D(glType,0,glFormat,width,height,0,hostFormat,dataType,data);
			GLHelper::checkGLError();
		} else if(type==CubeMap) {
    		if(data==0) {
				glTexImage2D(PositiveX,0,glFormat,width,height,0,hostFormat,dataType,0);
				glTexImage2D(NegativeX,0,glFormat,width,height,0,hostFormat,dataType,0);
				glTexImage2D(PositiveY,0,glFormat,width,height,0,hostFormat,dataType,0);
				glTexImage2D(NegativeY,0,glFormat,width,height,0,hostFormat,dataType,0);
				glTexImage2D(PositiveZ,0,glFormat,width,height,0,hostFormat,dataType,0);
				glTexImage2D(NegativeZ,0,glFormat,width,height,0,hostFormat,dataType,0);
    		} else {
				glTexImage2D(PositiveX,0,glFormat,width,height,0,hostFormat,dataType,((char *)data));
				glTexImage2D(NegativeX,0,glFormat,width,height,0,hostFormat,dataType,((char *)data)+imageSize);
				glTexImage2D(PositiveY,0,glFormat,width,height,0,hostFormat,dataType,((char *)data)+2*imageSize);
				glTexImage2D(NegativeY,0,glFormat,width,height,0,hostFormat,dataType,((char *)data)+3*imageSize);
				glTexImage2D(PositiveZ,0,glFormat,width,height,0,hostFormat,dataType,((char *)data)+4*imageSize);
				glTexImage2D(NegativeZ,0,glFormat,width,height,0,hostFormat,dataType,((char *)data)+5*imageSize);
    		}
		} else {
			// TODO texture buffer support
			glGenBuffers(1,&textureBufferID);
			glBindBuffer(GL_TEXTURE_BUFFER,textureBufferID);
			glBufferData(GL_TEXTURE_BUFFER,imageSize,data,GL_DYNAMIC_READ);
		}

		textureUnit=-1;
	}

	Texture::~Texture()
	{
		std::cout << "deleting texture " + name << std::endl;
		glDeleteTextures(1,&textureID);
		if(glIsTexture(textureID)!=GL_FALSE) {
			throw gltb::Exception("failed to delete texture " + name,"Texture::~Texture");
		}

		if(data) {
			delete[] data;
		}

		if(type==Buffer) {
			glDeleteBuffers(1,&textureBufferID);
		}
	}

	void Texture::bind(int textureUnit)
	{
		if(textureUnit!=activeTextureUnit) {
			glActiveTexture(GL_TEXTURE0+textureUnit);
			activeTextureUnit=textureUnit;
		}
		glEnable(glType);
		glBindTexture(glType,textureID);
		this->textureUnit=textureUnit;
	}

	void Texture::unbind()
	{
		if(textureUnit==-1) {
			return;
		}
		if(textureUnit!=activeTextureUnit) {
			glActiveTexture(GL_TEXTURE0+textureUnit);
			activeTextureUnit=textureUnit;
		}
		glDisable(glType);
		textureUnit=-1;
	}

	void Texture::setFilter(Filter filter)
	{
		bind(0);
		glTexParameteri(glType,GL_TEXTURE_MAG_FILTER, filter);
		glTexParameteri(glType,GL_TEXTURE_MIN_FILTER, filter);
		unbind();
	}

	void Texture::setWrapMode(WrapMode wrapMode)
	{
		bind(0);
		glTexParameteri(glType,GL_TEXTURE_WRAP_S,wrapMode);
		glTexParameteri(glType,GL_TEXTURE_WRAP_T,wrapMode);
		unbind();
	}

	void Texture::generateMipmap()
	{
		bind(0);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		glGenerateMipmap(glType);
		unbind();
	}

	void Texture::readBack(int regionX, int regionY, int regionWidth, int regionHeight)
	{
		if(regionHeight==-1) {
			regionX=regionY=0;
			regionWidth=width;
			regionHeight=height;
		}

		if(!data) {
			int numSides;
			switch(type) {
			case TwoD:
			case TwoDRect:
			case Buffer:
				numSides=1;
				break;
			case CubeMap:
				numSides=6;
				break;
			case ThreeD:
				numSides=depth;
			}

			int numTexels=numSides*width*height;

			data=new char[textureFormatDefinitions[format].bytesPerPixel*numTexels];
		}

		if(type!=Buffer) {
			if(readBackFB==0) {
				readBackFB=FramebufferManager::getInstance()->createFramebuffer(width,height,name + "_ReadBack");

				if(type==TwoD || type==TwoDRect) {
					readBackFB->attachTexture(Framebuffer::ColorAttachment,this);
				} else if(type==CubeMap) {
					readBackFB->attachTexture(Framebuffer::ColorAttachment,this,PositiveX);
				} else if(type==ThreeD) {
					readBackFB->attachTextureLayer(Framebuffer::ColorAttachment,this,0);
				}

				readBackFB->activate();

				if(!readBackFB->isValid()) {
					throw gltb::Exception("Texture readback failed - readback FBO not valid","Texture::readBack()");
				}
			}

			if(type==TwoD || type==TwoDRect) {
				readBackFB->activate();

				glReadPixels(regionX,regionY,regionWidth,regionHeight,hostFormat,dataType,data);

				readBackFB->deactivate();
			} else if(type==CubeMap) {
				Texture::CubeSide sides[6] = { PositiveX, NegativeX,
											   PositiveY, NegativeY,
											   PositiveZ, NegativeZ };
				for(int i=0;i<6;i++) {
					readBackFB->attachTexture(Framebuffer::ColorAttachment,gltb::RefPtr<Texture>(this),sides[i]);
					readBackFB->activate();
					glReadPixels(regionX,regionY,regionWidth,regionHeight,hostFormat,dataType,((char*)data)+i*imageSize);
					readBackFB->deactivate();
				}
			} else if(type==ThreeD) {
				for(int i=0;i<depth;i++) {
					readBackFB->attachTextureLayer(Framebuffer::ColorAttachment,this,i);
					readBackFB->activate();
					glReadPixels(regionX,regionY,regionWidth,regionHeight,hostFormat,dataType,((char*)data)+i*imageSize/depth);
					readBackFB->deactivate();
				}
			}
		} else {
			glBindBuffer(GL_TEXTURE_BUFFER,textureBufferID);
			void *temp=glMapBuffer(GL_TEXTURE_BUFFER,GL_READ_ONLY);
			memcpy(data,temp,imageSize);
			glUnmapBuffer(GL_TEXTURE_BUFFER);
		}
	}

	void Texture::reupload()
	{
		if(!data) {
			return;
		}

		if(type!=Buffer) {
			glBindTexture(glType,textureID);
			if(type==TwoD || type==TwoDRect) {
				glTexImage2D(glType,0,glFormat,width,height,0,hostFormat,dataType,data);
			} else if(type==CubeMap) {
				Texture::CubeSide sides[6] = { PositiveX, NegativeX,
											   PositiveY, NegativeY,
											   PositiveZ, NegativeZ };
				for(int i=0;i<6;i++) {
					glTexImage2D(sides[i],0,glFormat,width,height,0,hostFormat,dataType,((char*)data)+i*imageSize);
				}
			} else if(type==ThreeD) {
				glTexImage3D(glType,0,glFormat,width,height,depth,0,hostFormat,dataType,data);
			}
		} else {
			glBindBuffer(GL_TEXTURE_BUFFER,textureBufferID);
			void *temp=glMapBuffer(GL_TEXTURE_BUFFER,GL_WRITE_ONLY);
			memcpy(temp,data,imageSize);
			glUnmapBuffer(GL_TEXTURE_BUFFER);
		}
	}
}
