#include <iostream>

#include <GL/glew.h>

#include "GLTB/exception.h"

#include "GL3Support/glhelper.h"
#include "GL3Support/statemanager.h"
#include "GL3Support/vertexbuffer.h"

namespace GL3Support {
	VertexBuffer::VertexBuffer(void *sourceBuffer, int size, std::vector<ElementDescription> format, ExpectedUsage expectedUsage)
	{
		bufferOpened=false;
		data=sourceBuffer;

		state=new State();

		bufferFormat=format;
		validateFormat();
		glGenBuffers(1,&vertexBufferID);
		glBindBuffer(GL_ARRAY_BUFFER,vertexBufferID);
		glBufferData(GL_ARRAY_BUFFER,size,sourceBuffer,expectedUsage);
		glBindBuffer(GL_ARRAY_BUFFER,0);

		glGenVertexArrays(1,&vertexArrayID);
		vertexArraySetup=false;
	}

	VertexBuffer::~VertexBuffer()
	{
		if(data!=0) {
			delete data;
		}
		glDeleteBuffers(1,&vertexBufferID);
		glDeleteVertexArrays(1,&vertexArrayID);
	}

	void *VertexBuffer::getBuffer()
	{
		bufferOpened=true;
		glBindBuffer(GL_ARRAY_BUFFER,vertexBufferID);
		void *bufferData=glMapBuffer(GL_ARRAY_BUFFER,GL_READ_WRITE);
		glBindBuffer(GL_ARRAY_BUFFER,0);

		return bufferData;
	}

	void VertexBuffer::commitBuffer()
	{
		glBindBuffer(GL_ARRAY_BUFFER,vertexBufferID);
		glUnmapBuffer(GL_ARRAY_BUFFER);
		glBindBuffer(GL_ARRAY_BUFFER,0);
		bufferOpened=false;
	}

	void VertexBuffer::render(int first, int count)
	{
		if(bufferOpened) {
			throw gltb::Exception("cannot render an opened vertex buffer","GL3Support::VertexBuffer::render()");
		}

//		std::cout << "setting state for vertex buffer" << std::endl;
		if(state!=0) {
			StateManager::getInstance()->setState(state);
		}
//		GLHelper::checkGLError();

//		glEnable(GL_TEXTURE_2D);
//		glEnable(GL_TEXTURE_CUBE_MAP);

		// render vertex buffers
//		glEnable(GL_VERTEX_ARRAY);

		if(!vertexArraySetup) {
			setupVertexArray();
			vertexArraySetup=true;
		}

		glBindVertexArray(vertexArrayID);
		glBindBuffer(GL_ARRAY_BUFFER,vertexBufferID);

		if(indexed) {
			if(count==-1) {
				count=indexCount;
			}

			if(first<indexCount) {
				if(first+count>indexCount) {
					count=indexCount-first;
				}

				switch(primitiveType) {
				case Points:
					glDrawElements(GL_POINTS,count,GL_UNSIGNED_INT,(GLvoid*)(indexOffset+sizeof(unsigned int)*first));
					break;
				case Lines:
					glDrawElements(GL_LINES,count,GL_UNSIGNED_INT,(GLvoid*)(indexOffset+sizeof(unsigned int)*first));
					break;
				case LineStrip:
					glDrawElements(GL_LINE_STRIP,count,GL_UNSIGNED_INT,(GLvoid*)(indexOffset+sizeof(unsigned int)*first));
					break;
				case Triangles:
					glDrawElements(GL_TRIANGLES,count,GL_UNSIGNED_INT,(GLvoid*)(indexOffset+sizeof(unsigned int)*first));
					break;
				case TriangleStrip:
					glDrawElements(GL_TRIANGLE_STRIP,count,GL_UNSIGNED_INT,(GLvoid*)(indexOffset+sizeof(unsigned int)*first));
					break;
				case Quads:
					glDrawElements(GL_QUADS,count,GL_UNSIGNED_INT,(GLvoid*)(indexOffset+sizeof(unsigned int)*first));
					break;
				}
			}
		} else {
			if(count==-1) {
				count=vertexCount;
			}

			if(first<vertexCount) {
				if(first+count>vertexCount) {
					count=vertexCount-first;
				}

				switch(primitiveType) {
				case Points:
					glDrawArrays(GL_POINTS,first,count);
					break;
				case Lines:
					glDrawArrays(GL_LINES,first,count);
					break;
				case LineStrip:
					glDrawArrays(GL_LINE_STRIP,first,count);
					break;
				case Triangles:
					glDrawArrays(GL_TRIANGLES,first,count);
					break;
				case TriangleStrip:
					glDrawArrays(GL_TRIANGLE_STRIP,first,count);
					break;
				case Quads:
					glDrawArrays(GL_QUADS,first,count);
					break;
				}
			}
		}

		glBindBuffer(GL_ARRAY_BUFFER,0);
		glBindVertexArray(0);
//		glDisable(GL_VERTEX_ARRAY);
//		glDisable(GL_NORMAL_ARRAY);

//		glDisable(GL_TEXTURE_2D);
//		glDisable(GL_TEXTURE_CUBE_MAP);
	}

	void VertexBuffer::validateFormat()
	{
		bool indexBufferFound=false;
		indexed=false;
		indexCount=-1;
		vertexCount=-1;

		for(unsigned int i=0;i<bufferFormat.size();i++) {
//			std::cout << "validating vertex buffer element " << bufferFormat[i].name << std::endl;
			switch(bufferFormat[i].semantics) {
			case Index:
				if(indexBufferFound) {
					throw gltb::Exception("vertex buffer has more than one index element","GL3Support::VertexBuffer::validateFormat()");
				} else {
					indexBufferFound=true;
				}
				if(bufferFormat[i].type!=Int) {
					throw gltb::Exception("vertex buffer index element has unsupported type","GL3Support::VertexBuffer::validateFormat()");
				}
				if(bufferFormat[i].stride!=0) {
					throw gltb::Exception("vertex buffer index element must have a stride of 0","GL3Support::VertexBuffer::validateFormat()");
				}
				indexed=true;
				indexCount=bufferFormat[i].count;
				indexOffset=bufferFormat[i].offset;
				break;
			case Attribute:
				if(vertexCount==-1) {
					vertexCount=bufferFormat[i].count;
				} else if(vertexCount!=bufferFormat[i].count) {
					throw gltb::Exception("vertex buffer has inconsistent element counts","GL3Support::VertexBuffer::validateFormat()");
				}
				break;
			}
		}
	}

	void VertexBuffer::setupVertexArray()
	{
		glBindVertexArray(vertexArrayID);

		glBindBuffer(GL_ARRAY_BUFFER,vertexBufferID);

		for(unsigned int i=0;i<bufferFormat.size();i++) {
			switch(bufferFormat[i].semantics) {
			case Index:
				// TODO
				break;
			case Attribute:
				if(state->shader!=0) {
					int attributeIndex=glGetAttribLocation(state->shader->_getProgramID(),bufferFormat[i].name.c_str());

					if(attributeIndex==-1) {
						std::cout << "WARNING: expected attribute " << bufferFormat[i].name << " is not in shader " << state->shader->getDefinitionShaderName() << std::endl;
						continue;
					}

					glEnableVertexAttribArray(attributeIndex);

					switch(bufferFormat[i].type) {
					case Char:
						glVertexAttribPointer(attributeIndex,1,GL_BYTE,GL_FALSE,bufferFormat[i].stride,(GLvoid*)bufferFormat[i].offset);
						break;
					case Int:
						glVertexAttribIPointer(attributeIndex,1,GL_INT,bufferFormat[i].stride,(GLvoid*)bufferFormat[i].offset);
						break;
					case IVec2:
						glVertexAttribIPointer(attributeIndex,2,GL_INT,bufferFormat[i].stride,(GLvoid*)bufferFormat[i].offset);
						break;
					case IVec3:
						glVertexAttribIPointer(attributeIndex,3,GL_INT,bufferFormat[i].stride,(GLvoid*)bufferFormat[i].offset);
						break;
					case IVec4:
						glVertexAttribIPointer(attributeIndex,4,GL_INT,bufferFormat[i].stride,(GLvoid*)bufferFormat[i].offset);
						break;
					case Float:
						glVertexAttribPointer(attributeIndex,1,GL_FLOAT,GL_FALSE,bufferFormat[i].stride,(GLvoid*)bufferFormat[i].offset);
						break;
					case Vec2:
						glVertexAttribPointer(attributeIndex,2,GL_FLOAT,GL_FALSE,bufferFormat[i].stride,(GLvoid*)bufferFormat[i].offset);
						break;
					case Vec3:
						glVertexAttribPointer(attributeIndex,3,GL_FLOAT,GL_FALSE,bufferFormat[i].stride,(GLvoid*)bufferFormat[i].offset);
						break;
					case Vec4:
						glVertexAttribPointer(attributeIndex,4,GL_FLOAT,GL_FALSE,bufferFormat[i].stride,(GLvoid*)bufferFormat[i].offset);
						break;
					}
				}
				break;
			}
		}
		if(indexed) {
			glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,vertexBufferID);
		}

		glBindVertexArray(0);
	}
}
