#include "GL3Support/vertexbufferbuilder.h"

namespace GL3Support {
	VertexBufferBuilder::VertexBufferBuilder(VertexBuffer::PrimitiveType primitiveType)
      : primitiveType(primitiveType), 
		hasNormals(false),
		currentNormal(0,0,1),
		hasColours(false),
		currentColour(1,1,1,1),
		hasUVCoordinates(false),
		currentUVCoordinate(0,0),
		vertexAttributeName("vertices"), 
		normalAttributeName("normals"), 
		colourAttributeName("colours")
	{
	}

	void VertexBufferBuilder::setVertexAttributeName(std::string name) 
	{
		vertexAttributeName=name;
	}

	void VertexBufferBuilder::setNormalAttributeName(std::string name)
	{
		normalAttributeName=name;
	}

	void VertexBufferBuilder::setColourAttributeName(std::string name)
	{
		colourAttributeName=name;
	}

	void VertexBufferBuilder::setUVCoordinatesAttributeName(std::string name)
	{
		uvCoordinatesAttributeName=name;
	}

	void VertexBufferBuilder::vertex(glm::vec3 vertex)
	{
		vertices.push_back(vertex);
		if(hasNormals) {
			normals.push_back(currentNormal);
		}
		if(hasColours) {
			colours.push_back(currentColour);
		}
		if(hasUVCoordinates) {
			uvCoordinates.push_back(currentUVCoordinate);
		}
	}

	void VertexBufferBuilder::normal(glm::vec3 normal)
	{
		if(!hasNormals) {
			normals.resize(vertices.size());
			for(unsigned int i=0;i<normals.size();i++) {
				normals[i]=currentNormal;
			}
		}
		currentNormal=normal;
		hasNormals=true;
	}

	void VertexBufferBuilder::colour(glm::vec4 colour)
	{
		if(!hasColours) {
			colours.resize(vertices.size());
			for(unsigned int i=0;i<colours.size();i++) {
				colours[i]=currentColour;
			}
		}
		currentColour=colour;
		hasColours=true;
	}

	void VertexBufferBuilder::uvCoordinate(glm::vec2 uvCoordinate)
	{
		if(!hasUVCoordinates) {
			uvCoordinates.resize(vertices.size());
			for(unsigned int i=0;i<uvCoordinates.size();i++) {
				uvCoordinates[i]=currentUVCoordinate;
			}
		}
		currentUVCoordinate=uvCoordinate;
		hasUVCoordinates=true;
	}

	gltb::RefPtr<VertexBuffer> VertexBufferBuilder::generateVertexBuffer(VertexBuffer::ExpectedUsage expectedUsage)
	{
		// calculate buffer size and allocate buffer
		int size=3*sizeof(float)*vertices.size();
		if(hasNormals) {
			size+=3*sizeof(float)*normals.size();
		}
		if(hasColours) {
			size+=4*sizeof(float)*colours.size();
		}
		if(hasUVCoordinates) {
			size+=2*sizeof(float)*uvCoordinates.size();
		}

		float *bufferData=new float[size];

		// fill buffer with data
		int index=0;

		for(unsigned int i=0;i<vertices.size();i++) {
			bufferData[index++]=vertices[i].x;
			bufferData[index++]=vertices[i].y;
			bufferData[index++]=vertices[i].z;
		}
		if(hasNormals) {
			for(unsigned int i=0;i<normals.size();i++) {
				bufferData[index++]=normals[i].x;
				bufferData[index++]=normals[i].y;
				bufferData[index++]=normals[i].z;
			}
		}
		if(hasColours) {
			for(unsigned int i=0;i<colours.size();i++) {
				bufferData[index++]=colours[i].r;
				bufferData[index++]=colours[i].g;
				bufferData[index++]=colours[i].b;
				bufferData[index++]=colours[i].a;
			}
		}
		if(hasUVCoordinates) {
			for(unsigned int i=0;i<uvCoordinates.size();i++) {
				bufferData[index++]=uvCoordinates[i].x;
				bufferData[index++]=uvCoordinates[i].y;
			}
		}

		// create buffer format description
		std::vector<VertexBuffer::ElementDescription> format;

		VertexBuffer::ElementDescription element;
		element.semantics=VertexBuffer::Attribute;
		element.type=VertexBuffer::Vec3;
		element.name=vertexAttributeName;
		element.offset=0;
		element.stride=0;
		element.count=vertices.size();
		format.push_back(element);
		
		element.offset=3*sizeof(float)*vertices.size();

		if(hasNormals) {
			element.name=normalAttributeName;
			format.push_back(element);
			element.offset+=3*sizeof(float)*vertices.size();
		}
		if(hasColours) {
			element.name=colourAttributeName;
			element.type=VertexBuffer::Vec4;
			format.push_back(element);
			element.offset+=4*sizeof(float)*vertices.size();
		}
		if(hasUVCoordinates) {
			element.name=uvCoordinatesAttributeName;
			element.type=VertexBuffer::Vec2;
			format.push_back(element);
		}

		gltb::RefPtr<VertexBuffer> buffer=new VertexBuffer(bufferData,size,format,expectedUsage);
		buffer->setPrimitiveType(primitiveType);

		return buffer;
	}
}
