#include <fstream>
#include <iomanip>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include <cmath>

#include <FreeImage.h>

#include <GL/glew.h>
#include <GLFW/glfw3.h>

#include <glm/glm.hpp>

#include <GLTB/exception.h>
#include <GLTB/rng.h>
#include <GLTB/geom/triangle.h>

#include <GL3Support/framebuffer.h>
#include <GL3Support/framebuffermanager.h>
#include <GL3Support/glhelper.h>
#include <GL3Support/texture.h>

int imageWidth, imageHeight;
unsigned char *origImage;

void loadInput()
{
  FREE_IMAGE_FORMAT format = FreeImage_GetFileType("input.png",0);
  FIBITMAP* image = FreeImage_Load(format, "input.png");
  
  imageWidth = FreeImage_GetWidth(image);
  imageHeight = FreeImage_GetHeight(image);

  FIBITMAP *converted = FreeImage_ConvertTo32Bits(image);
  
  unsigned char* pixels = (unsigned char*)FreeImage_GetBits(converted);

  origImage = new unsigned char[4 * imageWidth * imageHeight];
  for(int j = 0; j < imageWidth * imageHeight; j++){
    origImage[4 * j + 0]= pixels[4 * j + 2];
    origImage[4 * j + 1]= pixels[4 * j + 1];
    origImage[4 * j + 2]= pixels[4 * j + 0];
    origImage[4 * j + 3]= pixels[4 * j + 3];
  }

  FreeImage_Unload(image);
  FreeImage_Unload(converted);
}

gltb::RefPtr<GL3Support::Texture> resultTexture;
gltb::RefPtr<GL3Support::Framebuffer> resultFramebuffer;

void initTexture()
{
  resultTexture = new GL3Support::Texture(GL3Support::Texture::TwoD, GL3Support::Texture::UCharRGBA, imageWidth, imageHeight, "resultTexture");
  
  resultFramebuffer = GL3Support::FramebufferManager::getInstance()->createFramebuffer(imageWidth, imageHeight, "resultFramebuffer");
  resultFramebuffer->attachTexture(GL3Support::Framebuffer::ColorAttachment, resultTexture);
  resultFramebuffer->activate();
  if(!resultFramebuffer->isValid()) {
    throw gltb::Error("framebuffer is invalid", "initTexture()");
  }
  resultFramebuffer->deactivate();
}

const int numTriangles = 200;
gltb::Random rng;

std::vector<float> triangleData;

std::vector<float> generateInitialTriangleData(int numTriangles)
{
	std::vector<float> triangleData;

	triangleData.resize(6 * 3 * numTriangles);
	for(size_t i = 0; i < triangleData.size(); i++) {
		triangleData[i] = rng.genrandReal1();
	}
  
	for(int index = 0; index < 3 * numTriangles; index++) {
		int x = imageWidth * triangleData[6 * index + 4];
		int y = imageHeight * triangleData[6 * index + 5];
		
		triangleData[6 * index + 0] = origImage[4 * (y * imageWidth + x) + 0] / 256.0;
		triangleData[6 * index + 1] = origImage[4 * (y * imageWidth + x) + 1] / 256.0;
		triangleData[6 * index + 2] = origImage[4 * (y * imageWidth + x) + 2] / 256.0;
	}

  
	return triangleData;
}

void renderTriangles(const std::vector<float> &triangleData)
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glViewport(0, 0, imageWidth, imageHeight);
  
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0, 1, 0, 1, -1.0, 1.0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  
//   glColor3f(1, 1, 1);
//   glBegin(GL_TRIANGLES);
//   glVertex3f(0, 0, 0);
//   glVertex3f(100, 0, 0);
//   glVertex3f(0, 100, 0);
//   glEnd();
  
  // old style fixed function pipeline will do nicely here
  glBegin(GL_TRIANGLES);
  for(size_t i = 0; i < triangleData.size(); i += 6) {
    glColor4f(triangleData[i], triangleData[i + 1], triangleData[i + 2], triangleData[i + 3]);
    glVertex3f(triangleData[i + 4], triangleData[i + 5], 0.0f);
  }  
  glEnd();

  GL3Support::GLHelper::checkGLError();  
}

void renderTrianglesToTexture(const std::vector<float> &triangleData)
{
  // activate texture framebuffer here for rendering
  resultFramebuffer->activate();
  renderTriangles(triangleData);
  resultFramebuffer->deactivate();
}

float imageError(const std::vector<float> &triangleData)
{
	float error = 0;

	resultTexture->readBack();
	const unsigned char *data = static_cast<const unsigned char*>(resultTexture->getData());

	int pixelStride = 4;
	for(int y = 0; y < imageHeight; y++) {  
		for(int x = 0; x < imageWidth; x++) {
			int i = y * imageWidth + x;
			for(int j = 0; j < 3; j++) {
				float pixelError = abs(data[i * pixelStride + j] - origImage[i * pixelStride + j]);
				
				// punish gradients not preset in original image
// 				float gradientError = 0;
// 				if(x > 0 && x < imageWidth - 1 && y > 0 && y < imageHeight - 1) {
// 					gradientError += abs(abs(data[(i - 1) * pixelStride + j] - data[i * pixelStride + j]) - abs(origImage[(i - 1) * pixelStride + j] - origImage[i * pixelStride + j]));
// 					gradientError += abs(abs(data[(i + 1) * pixelStride + j] - data[i * pixelStride + j]) - abs(origImage[(i + 1) * pixelStride + j] - origImage[i * pixelStride + j]));
// 					gradientError += abs(abs(data[(i - imageWidth) * pixelStride + j] - data[i * pixelStride + j]) - abs(origImage[(i - imageWidth) * pixelStride + j] - origImage[i * pixelStride + j]));
// 					gradientError += abs(abs(data[(i + imageWidth) * pixelStride + j] - data[i * pixelStride + j]) - abs(origImage[(i + imageWidth) * pixelStride + j] - origImage[i * pixelStride + j]));
// 				}
// 				pixelError += 0.25 * gradientError;
				
				error += pixelError * pixelError;
			}
		}
	}
	
	// punish large total triangle area and slivers (high ratio of edge length to area)
	float area = 0;
	float edgeRatio;
	for(size_t i = 0; i < triangleData.size(); i += 18) {
		gltb::Triangle<glm::vec3> t(glm::vec3(triangleData[i + 4], triangleData[i + 5], 0.0f), glm::vec3(triangleData[i + 10], triangleData[i + 11], 0.0f), glm::vec3(triangleData[i + 16], triangleData[i + 17], 0.0f));
		float triangleArea = t.getArea();; 
		area += triangleArea;
		float circumference = 0;
		circumference += glm::length(t.b - t.a);
		circumference += glm::length(t.c - t.a);
		circumference += glm::length(t.c - t.b);
		edgeRatio += circumference / area;
	}
	error += 10000 * area;
	error += 10 * edgeRatio;
	
	return sqrt(error / (imageWidth * imageHeight));
}

std::vector<float> mutateTriangles(const std::vector<float> &triangleData, size_t iteration)
{
// 	const float mutationScale = 0.1 + 0.9 * (1 - exp(-(iteration / 10000.0)));
// 	if(iteration % 1000 == 0) {
// 		std::cout << iteration << " " << mutationScale << std::endl;
// 	}
	const float mutationScale = 0.1;
	std::vector<float> mutated = triangleData;
  
	// mutate triangle data randomly
	int numMutations = 0.5 * triangleData.size() * rng.genrandReal1();

	for(int i = 0; i < numMutations; i++) {
		if(rng.genrandReal1() < 0.9) {
			// mutate position
			int index = (triangleData.size() / 6) * rng.genrandReal1();
			int x = imageWidth * mutated[6 * index + 4];
			int y = imageHeight * mutated[6 * index + 5];
			
			mutated[6 * index + 0] = origImage[4 * (y * imageWidth + x) + 0] / 256.0;
			mutated[6 * index + 1] = origImage[4 * (y * imageWidth + x) + 1] / 256.0;
			mutated[6 * index + 2] = origImage[4 * (y * imageWidth + x) + 2] / 256.0;
		} else {
			// mutate color
//			int index = (triangleData.size() / 6) * rng.genrandReal1();
//			index = index * 6 + 4 * rng.genrandReal1();
			int index = triangleData.size() * rng.genrandReal1();
			mutated[index] += mutationScale * (2 * rng.genrandReal1() - 1);
			if(mutated[index] > 1) {
				mutated[index] = 1;
				//mutated[index] = triangleData[index];
			}
			if(mutated[index] < 0) {
				mutated[index] = 0;
				//mutated[index] = triangleData[index];
			}
		}
// 		int index = (triangleData.size() / 6) * rng.genrandReal1();
// 		mutated[6 * index + 4] = rng.genrandReal1();
// 		mutated[6 * index + 5] = rng.genrandReal1();
// 		int x = imageWidth * mutated[6 * index + 4];
// 		int y = imageHeight * mutated[6 * index + 5];
// 		
// 		mutated[6 * index + 0] = origImage[4 * (y * imageWidth + x) + 0] / 256.0;
// 		mutated[6 * index + 1] = origImage[4 * (y * imageWidth + x) + 1] / 256.0;
// 		mutated[6 * index + 2] = origImage[4 * (y * imageWidth + x) + 2] / 256.0;
// 		//mutated[6 * index + 3] = 1;
	}


	return mutated;
}

void renderTrianglesToScreen(const std::vector<float> &triangleData, GLFWwindow *window)
{
  renderTriangles(triangleData);
  glfwSwapBuffers(window);
}

void savePPM(std::string filename)
{
  std::ofstream output(filename);
  
  output << "P6\n";
  output << imageWidth << " " << imageHeight << " 255\n";
  
  // Write actual image data from texture
  resultTexture->readBack();
  const char *data = static_cast<const char*>(resultTexture->getData());
  //const char *data = (char*) origImage;
  for(int i = 0; i < imageWidth * imageHeight; i++) {
    output.write(&data[4 * i], 3 * sizeof(char));
  }
  
  output.close();
}

int main(int argc, char *argv[])
{
  try {
    loadInput();
    
    // initialize GLFW
    glfwInit();
    glfwWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, GL_TRUE);
    glfwWindowHint(GLFW_SAMPLES, 4);
    GLFWwindow *window = glfwCreateWindow(imageWidth, imageHeight, "vertexcolor", NULL, NULL);
    glfwShowWindow(window);
    glfwMakeContextCurrent(window);
    if(glewInit() != GLEW_OK) {
      throw gltb::Error("glew init failed", "main()");
    }
    
    GL3Support::GLHelper::enableDebugLogging();
    GL3Support::GLHelper::checkGLError();
    
    initTexture();
    
    std::vector<float> bestTriangles = generateInitialTriangleData(numTriangles);
    renderTrianglesToTexture(bestTriangles);
    float bestError = imageError(bestTriangles);
    std::cout << "initial error: " << bestError << std::endl;
    renderTrianglesToScreen(bestTriangles, window);
    
    size_t iteration = 0;
    bool doContinue = true;
    while(doContinue) {
      // run optimization
      std::vector<float> newTriangles = mutateTriangles(bestTriangles, iteration);
      renderTrianglesToTexture(newTriangles);
      float newError = imageError(newTriangles);
      if(newError < bestError) {
        bestError = newError;
        bestTriangles = newTriangles;
        std::cout << "new best candidate: " << newError << std::endl;
	renderTrianglesToScreen(bestTriangles, window);
      }
      
      iteration++;
      if(iteration % 10000 == 0) {
	renderTrianglesToTexture(bestTriangles);
        std::ostringstream filename;
        filename << "iteration-" << std::setfill('0') << std::setw(6) << iteration / 10000 << ".ppm";
        std::cout << "saving " << filename.str() << std::endl;
        savePPM(filename.str());
      }
      
      if(glfwGetKey(window, GLFW_KEY_ESCAPE) == GLFW_PRESS) {
	doContinue = false;
      }
    }
    
    glfwTerminate();
  } catch(gltb::Exception &e) {
    std::cerr << e.asFormattedText() << std::endl;
  }
  
  return 0;
}
