#include <algorithm>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include <cmath>

#include <FreeImage.h>

#include <GL/glew.h>
#include <GLFW/glfw3.h>

#include <glm/glm.hpp>

#include <GLTB/exception.h>
#include <GLTB/rng.h>
#include <GLTB/geom/triangle.h>

#include <GL3Support/framebuffer.h>
#include <GL3Support/framebuffermanager.h>
#include <GL3Support/glhelper.h>
#include <GL3Support/texture.h>

int imageWidth, imageHeight;
unsigned char *origImage;

void loadInput()
{
  FREE_IMAGE_FORMAT format = FreeImage_GetFileType("input.png",0);
  FIBITMAP* image = FreeImage_Load(format, "input.png");
  
  imageWidth = FreeImage_GetWidth(image);
  imageHeight = FreeImage_GetHeight(image);

  FIBITMAP *converted = FreeImage_ConvertTo32Bits(image);
  
  unsigned char* pixels = (unsigned char*)FreeImage_GetBits(converted);

  origImage = new unsigned char[4 * imageWidth * imageHeight];
  for(int j = 0; j < imageWidth * imageHeight; j++){
    origImage[4 * j + 0]= pixels[4 * j + 2];
    origImage[4 * j + 1]= pixels[4 * j + 1];
    origImage[4 * j + 2]= pixels[4 * j + 0];
    origImage[4 * j + 3]= pixels[4 * j + 3];
  }

  FreeImage_Unload(image);
  FreeImage_Unload(converted);
}

gltb::RefPtr<GL3Support::Texture> resultTexture;
gltb::RefPtr<GL3Support::Framebuffer> resultFramebuffer;

void initTexture()
{
	resultTexture = new GL3Support::Texture(GL3Support::Texture::TwoD, GL3Support::Texture::UCharRGBA, imageWidth, imageHeight, "resultTexture");

	resultFramebuffer = GL3Support::FramebufferManager::getInstance()->createFramebuffer(imageWidth, imageHeight, "resultFramebuffer");
	resultFramebuffer->attachTexture(GL3Support::Framebuffer::ColorAttachment, resultTexture);
	resultFramebuffer->activate();
	if(!resultFramebuffer->isValid()) {
		throw gltb::Error("framebuffer is invalid", "initTexture()");
	}
	resultFramebuffer->deactivate();
}

const int gridWidth = 40;
const int gridHeight = 40;
gltb::Random rng;

struct Grid {
	std::vector<glm::vec3> colors;
	std::vector<glm::vec2> positions;
	std::vector<bool> flipped;
};

void generateInitialGrid(Grid &grid)
{
	grid.colors.resize(gridWidth * gridHeight);
	grid.positions.resize(gridWidth * gridHeight);
	grid.flipped.resize((gridWidth - 1) * (gridHeight - 1));

	for(int x = 0; x < gridWidth; x++) {
		for(int y = 0; y < gridHeight; y++) {
			int i = y * gridWidth + x;
			grid.positions[i].x = x / float(gridWidth - 1);
			grid.positions[i].y = y / float(gridHeight - 1);
			
			int imageX = std::min(int(grid.positions[i].x * imageWidth), imageWidth - 1);
			int imageY = std::min(int(grid.positions[i].y * imageHeight), imageHeight - 1);
			int pixel = imageY * imageWidth + imageX;
			
			grid.colors[i].r = origImage[4 * pixel + 0] / 256.0;
			grid.colors[i].g = origImage[4 * pixel + 0] / 256.0;
			grid.colors[i].b = origImage[4 * pixel + 0] / 256.0;
		}
	}
	for(size_t i = 0; i < grid.flipped.size(); i++) {
		grid.flipped[i] = false;
	}
}

void renderGrid(const Grid &grid)
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glViewport(0, 0, imageWidth, imageHeight);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, 1, 0, 1, -1.0, 1.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// old style fixed function pipeline will do nicely here
	glBegin(GL_TRIANGLES);
	int i;
	for(int x = 0; x < gridWidth - 1; x++) {
		for(int y = 0; y < gridHeight - 1; y++) {
			if(!grid.flipped[y * (gridWidth -1) + x]) {
				i = y * gridWidth + x;
				glColor3f(grid.colors[i].r, grid.colors[i].g, grid.colors[i].b);
				glVertex3f(grid.positions[i].x, grid.positions[i].y, 0.0f);
				i = y * gridWidth + x + 1;
				glColor3f(grid.colors[i].r, grid.colors[i].g, grid.colors[i].b);
				glVertex3f(grid.positions[i].x, grid.positions[i].y, 0.0f);
				i = (y + 1) * gridWidth + x;
				glColor3f(grid.colors[i].r, grid.colors[i].g, grid.colors[i].b);
				glVertex3f(grid.positions[i].x, grid.positions[i].y, 0.0f);

				i = (y + 1) * gridWidth + x;
				glColor3f(grid.colors[i].r, grid.colors[i].g, grid.colors[i].b);
				glVertex3f(grid.positions[i].x, grid.positions[i].y, 0.0f);
				i = y * gridWidth + x + 1;
				glColor3f(grid.colors[i].r, grid.colors[i].g, grid.colors[i].b);
				glVertex3f(grid.positions[i].x, grid.positions[i].y, 0.0f);
				i = (y + 1) * gridWidth + x + 1;
				glColor3f(grid.colors[i].r, grid.colors[i].g, grid.colors[i].b);
				glVertex3f(grid.positions[i].x, grid.positions[i].y, 0.0f);
			} else {
				i = y * gridWidth + x;
				glColor3f(grid.colors[i].r, grid.colors[i].g, grid.colors[i].b);
				glVertex3f(grid.positions[i].x, grid.positions[i].y, 0.0f);
				i = y * gridWidth + x + 1;
				glColor3f(grid.colors[i].r, grid.colors[i].g, grid.colors[i].b);
				glVertex3f(grid.positions[i].x, grid.positions[i].y, 0.0f);
				i = (y + 1) * gridWidth + x + 1;
				glColor3f(grid.colors[i].r, grid.colors[i].g, grid.colors[i].b);
				glVertex3f(grid.positions[i].x, grid.positions[i].y, 0.0f);

				i = (y + 1) * gridWidth + x;
				glColor3f(grid.colors[i].r, grid.colors[i].g, grid.colors[i].b);
				glVertex3f(grid.positions[i].x, grid.positions[i].y, 0.0f);
				i = y * gridWidth + x;
				glColor3f(grid.colors[i].r, grid.colors[i].g, grid.colors[i].b);
				glVertex3f(grid.positions[i].x, grid.positions[i].y, 0.0f);
				i = (y + 1) * gridWidth + x + 1;
				glColor3f(grid.colors[i].r, grid.colors[i].g, grid.colors[i].b);
				glVertex3f(grid.positions[i].x, grid.positions[i].y, 0.0f);
			}
		}
	}
	glEnd();

	GL3Support::GLHelper::checkGLError();  
}

void renderGridToTexture(const Grid &grid)
{
	// activate texture framebuffer here for rendering
	resultFramebuffer->activate();
	renderGrid(grid);
	resultFramebuffer->deactivate();
}

void renderGridToScreen(const Grid &grid, GLFWwindow *window)
{
	//glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	renderGrid(grid);
	//glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	glfwSwapBuffers(window);
}

float imageError()
{
	float error = 0;

	resultTexture->readBack();
	const unsigned char *data = static_cast<const unsigned char*>(resultTexture->getData());

	int pixelStride = 4;
	for(int y = 0; y < imageHeight; y++) {  
		for(int x = 0; x < imageWidth; x++) {
			int i = y * imageWidth + x;
			for(int j = 0; j < 3; j++) {
				float pixelError = abs(data[i * pixelStride + j] - origImage[i * pixelStride + j]);
				
				// punish gradients not preset in original image
// 				float gradientError = 0;
// 				if(x > 0 && x < imageWidth - 1 && y > 0 && y < imageHeight - 1) {
// 					gradientError += abs(abs(data[(i - 1) * pixelStride + j] - data[i * pixelStride + j]) - abs(origImage[(i - 1) * pixelStride + j] - origImage[i * pixelStride + j]));
// 					gradientError += abs(abs(data[(i + 1) * pixelStride + j] - data[i * pixelStride + j]) - abs(origImage[(i + 1) * pixelStride + j] - origImage[i * pixelStride + j]));
// 					gradientError += abs(abs(data[(i - imageWidth) * pixelStride + j] - data[i * pixelStride + j]) - abs(origImage[(i - imageWidth) * pixelStride + j] - origImage[i * pixelStride + j]));
// 					gradientError += abs(abs(data[(i + imageWidth) * pixelStride + j] - data[i * pixelStride + j]) - abs(origImage[(i + imageWidth) * pixelStride + j] - origImage[i * pixelStride + j]));
// 				}
// 				pixelError += 0.25 * gradientError;
				
				error += pixelError * pixelError;
			}
		}
	}
	
	return sqrt(error / (imageWidth * imageHeight));
}

void mutateGrid(Grid &grid)
{
	int numMutations = grid.positions.size() * rng.genrandReal1();

	for(int i = 0; i < numMutations; i++) {
 		if(rng.genrandReal1() < 0.9) {
			// mutate position
			int index = grid.positions.size() * rng.genrandReal1();
			
			int gridX = index % gridWidth;
			int gridY = index / gridWidth;
			
			if(gridX != 0 && gridX != gridWidth -1) {
				float center = 0.5 * (grid.positions[index + 1].x + grid.positions[index - 1].x);
				float range = grid.positions[index + 1].x - grid.positions[index - 1].x;
				grid.positions[index].x = center + 0.5 * range * (rng.genrandReal1() - 0.5);
			}
			if(gridY != 0 && gridY != gridHeight -1) {
				float center = 0.5 * (grid.positions[index + gridWidth].y + grid.positions[index - gridWidth].y);
				float range = grid.positions[index + gridWidth].y - grid.positions[index - gridWidth].y;
				grid.positions[index].y = center + 0.5 * range * (rng.genrandReal1() - 0.5);
			}
			
			int x = std::min(int(imageWidth * grid.positions[index].x), imageWidth - 1);
			int y = std::min(int(imageHeight * grid.positions[index].y), imageHeight -1);
			
			grid.colors[index].r = origImage[4 * (y * imageWidth + x) + 0] / 256.0;
			grid.colors[index].g = origImage[4 * (y * imageWidth + x) + 1] / 256.0;
			grid.colors[index].b = origImage[4 * (y * imageWidth + x) + 2] / 256.0;
		} else {
			int index = grid.flipped.size() * rng.genrandReal1();
			grid.flipped[index] = !grid.flipped[index];
		}
	}
}

void savePPM(std::string filename)
{
  std::ofstream output(filename);
  
  output << "P6\n";
  output << imageWidth << " " << imageHeight << " 255\n";
  
  // Write actual image data from texture
  resultTexture->readBack();
  const char *data = static_cast<const char*>(resultTexture->getData());
  //const char *data = (char*) origImage;
  for(int i = 0; i < imageWidth * imageHeight; i++) {
    output.write(&data[4 * i], 3 * sizeof(char));
  }
  
  output.close();
}

int main(int argc, char *argv[])
{
	try {
		loadInput();

		// initialize GLFW
		glfwInit();
		glfwWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, GL_TRUE);
		glfwWindowHint(GLFW_SAMPLES, 4);
		GLFWwindow *window = glfwCreateWindow(imageWidth, imageHeight, "vertexgrid", NULL, NULL);
		glfwShowWindow(window);
		glfwMakeContextCurrent(window);
		if(glewInit() != GLEW_OK) {
		throw gltb::Error("glew init failed", "main()");
		}

		GL3Support::GLHelper::enableDebugLogging();
		GL3Support::GLHelper::checkGLError();

		initTexture();

		Grid bestGrid;
		generateInitialGrid(bestGrid);
		renderGridToTexture(bestGrid);
		float bestError = imageError();
		std::cout << "initial error: " << bestError << std::endl;
		renderGridToScreen(bestGrid, window);
		{
//			glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
			renderGridToTexture(bestGrid);
			savePPM("initial.ppm");
//			return 0;
		}

		size_t iteration = 0;
		bool doContinue = true;
		while(doContinue) {
			// run optimization
			Grid newGrid = bestGrid;
			mutateGrid(newGrid);
			renderGridToTexture(newGrid);
			float newError = imageError();
			if(newError < bestError) {
				bestError = newError;
				bestGrid = newGrid;
				std::cout << "new best candidate: " << newError << std::endl;
				renderGridToScreen(bestGrid, window);
			}
      
			iteration++;
			if(iteration % 10000 == 0) {
				renderGridToTexture(bestGrid);
				std::ostringstream filename;
				filename << "iteration-" << std::setfill('0') << std::setw(6) << iteration / 10000 << ".ppm";
				std::cout << "saving " << filename.str() << std::endl;
				savePPM(filename.str());
			}

			if(glfwGetKey(window, GLFW_KEY_ESCAPE) == GLFW_PRESS) {
				doContinue = false;
			}
		}

		glfwTerminate();
	} catch(gltb::Exception &e) {
		std::cerr << e.asFormattedText() << std::endl;
	}
  
	return 0;
}
